;; new completion style  -*- lexical-binding:t -*-

;; There are four kinds of strings involved in completion. For most
;; categories, they are all the same; the following table describes
;; them for the absfile category.
;;
;; - user input
;;
;;   The string typed by the user in the minibuffer.
;;
;;   Also returned by try-completions when it extends the string.
;;
;;   For the absfile category, this is a partial file name plus any
;;   required uniquifying directories, formatted according to
;;   `find-absfile-uniquify-style'
;;
;; - completion table input
;;
;;   The string input to the completion table function.
;;
;;   The completion-try-completion, completion-all-completion, and
;;   test-completion functions must convert the user input string to a
;;   completion table input string when calling the corresponding
;;   `*-completion'.
;;
;;   For the absfile category, if the user input string contains any
;;   directory portion, or if action is 'lambda (called from
;;   test-completion), this contains the complete absolute directory
;;   name followed by the partial file name. Otherwise it is just the
;;   partial file name.
;;
;; - display
;;
;;   The strings returned by the completion-all-completions
;;   function. The displayed completion list consists of these strings
;;   with the common prefix deleted.
;;
;;   For the absfile category, this is the complete file name plus
;;   any required uniquifying directories, formatted according to
;;   find-absfile-uniquify-style.
;;
;; - data
;;
;;   The string returned by completion-get-data-string. In the future,
;;   this may be the string returned by completing-read; for now,
;;   completing-read returns a display string, and the caller must
;;   call completion-get-data-string.
;;
;;   The completion table returns a list of these strings when action
;;   is t.
;;
;;   For the absfile category, this is an absolute file name.
;;
;;
;; Currently, completion-try-completion, completion-all-completion
;; support style-specific implementations via completion-style-alist,
;; but test-completion and completion-get-data-string do
;; not. Style-specific code is added to test-completion in advice, and
;; in completion-get-data-string in a case statement.
;;
;; We advise completing-read-default to call
;; completion-get-data-string on the result string. That should be
;; called from completion--complete-and-exit, to replace the buffer
;; string either after the case fixup. That's not possible with
;; advice, because completion--complete-and-exit does not exit
;; normally; it throws a signal (FIXME: can use condition-case in
;; around advice?).


(defvar find-absfile-uniquify-style 'abbrev
  "Style used to format uniquifying directories.
One of:
- 'abbrev : minimal directories required to identify a unique file (may be empty)
- 'full   : absolute directory path or empty")

(defconst find-absfile-uniquify-regexp "^\\(.*\\)<\\(.*\\)>"
  "Regexp matching uniqufied file name.
Match 1 is the filename, match 2 is the relative directory.")

(defun find-absfile-uniquify-conflicts (conflicts)
  "Subroutine of `find-file-uniquify'."
  (let ((common-root ;; shared prefix of dirs in conflicts - may be nil
	 (fill-common-string-prefix (file-name-directory (nth 0 conflicts)) (file-name-directory (nth 1 conflicts)))))

    (let ((temp (cddr conflicts)))
      (while (and common-root
		  temp)
	(setq common-root (fill-common-string-prefix common-root (file-name-directory (pop temp))))))

    (when common-root
      ;; Trim `common-root' back to last '/'
      (let ((i (1- (length common-root))))
	(while (and (> i 0)
		    (not (= (aref common-root i) ?/)))
	  (setq i (1- i)))
	(setq common-root (substring common-root 0 (1+ i)))))

    (cl-mapcar
     (lambda (name)
       (concat (file-name-nondirectory name)
	       "<" (substring (file-name-directory name) (length common-root)) ">"))
     conflicts)
    ))

(defun find-absfile-uniquify (names)
  "Return a uniquified list of names built from NAMES.
NAMES contains absolute file names.

The result contains non-directory filenames with partial directory paths
appended."
  (cl-ecase find-absfile-uniquify-style
    (abbrev
     (let (result
	   conflicts ;; list of names where all non-directory names are the same.
	   )

       ;; Sort names so duplicates are grouped together
       (setq names (sort names (lambda (a b)
				 (string< (file-name-nondirectory a) (file-name-nondirectory b)))))

       (while names
	 (setq conflicts (list (pop names)))
	 (while (and names
		     (string= (file-name-nondirectory (car conflicts)) (file-name-nondirectory (car names))))
	     (push (pop names) conflicts))

	 (if (= 1 (length conflicts))
	     (push (file-name-nondirectory (car conflicts)) result)
	   (setq result (append (find-absfile-uniquify-conflicts conflicts) result)))
	 )
       (nreverse result)
       ))

    (full
     names)
    ))

(defun find-absfile-deuniquify (path name &optional force)
  "Convert display string NAME to table input string.
PATH is used to fill in missing directories.  If FORCE is
non-nil (default nil), NAME must exist somewhere in PATH; always
return absolute filename."
  (let ((match (string-match find-absfile-uniquify-regexp name)))
    (if (or match force)

	(let ((dirname  (or (when match (match-string 2 name)) ""))
	      (filename (or (when match (match-string 1 name)) name))
	      (temp-path path)
	      dir
	      matching-dir)

	  (cond
	   ((and (not force)
		 (= 0 (length dirname)))
	    filename)

	   ((or force
		(< 0 (length dirname)))
	    ;; Extend dirname to absolute directory name
	    (while (and temp-path
			(not matching-dir))
	      ;; `path' should only have directories (ie end in slash), but
	      ;; since it is often built by hand, it may not.
	      (setq dir (file-name-as-directory (pop temp-path)))

	      ;; FIXME: not doing completion on partial directory names yet.
	      (cond
	       (match
		(when (string-match dirname dir)
		  (setq matching-dir dir)))

	       (force
		(when (file-exists-p (concat dir name));; FIXME: Requires a flat path or a recursive path iterator
		  (setq matching-dir dir)))
	      )))
	   )
	  (concat matching-dir filename))

      ;; else not uniquified
      name)))

(defun completion-uniquify-absfile-try-completion (string table pred point)
  "Implement `completion-try-completion' for uniquify-absfile."
  ;; Trivial try-completion backend; don't handle boundaries yet
  (let* ((beforepoint (substring string 0 point))
	 (completion (try-completion (completion-absfile-deuniquify beforepoint table pred) table pred)))

    (if (stringp completion)
	;; multiple matches
	(cons completion (length completion))

      ;; either no match or one match
      completion)
    ))

(defun completion-uniquify-absfile-all-completions (string table pred point)
  "Implement `completion-all-completions' for uniquify-absfile."
  ;; Trivial all-completion backend; don't handle boundaries yet
  (find-absfile-uniquify
   (all-completions
    (completion-absfile-deuniquify (substring string 0 point) table pred)
    table
    pred)))

(defun completion-absfile-deuniquify (display-string table pred &optional force)
  "Implement `completion-get-data-string' for uniquify-absfile."
  (find-absfile-deuniquify (completion-metadata-get (completion-metadata display-string table pred) 'path)
			   display-string
			   force))

(defun completion-uniquify-absfile-get-data-string (display-string table pred)
  "Implement `completion-get-data-string' for uniquify-absfile."
  (completion-absfile-deuniquify display-string table pred t))

(defun completion-get-data-string (display-string table pred)
  "Return the data string corresponding to DISPLAY-STRING."
  ;; FIXME: should dispatch via completion-styles-alist
  ;; FIXME: should be called from completing-read somewhere
  (case (completion-metadata-get (completion-metadata display-string table pred) 'category)
    (absfile (completion-uniquify-absfile-get-data-string display-string table pred))
    (t display-string)
  ))

(defun find-absfile-test-completion-advice (orig-fun string table &optional pred)
  "Advice for `test-completion'; convert display string to table input."
  (let ((metadata (completion-metadata string table pred)))
    (case (completion-metadata-get metadata 'category)
      (absfile
       (setq string (find-absfile-deuniquify (completion-metadata-get metadata 'path) string t)))
      (t nil)
      ))
  (funcall orig-fun string table pred))

;; FIXME: using :filter-args (with a suitably modified advice function) gives an error about argument count
(advice-add #'test-completion :around #'find-absfile-test-completion-advice)

(defun find-absfile-completing-read-default-advice (orig-fun prompt collection &optional predicate
							     require-match initial-input hist def
							     inherit-input-method)
  "Advice for `completing-read-default'; convert display string to data string."
  (let ((display-string (funcall orig-fun prompt collection
				 predicate require-match initial-input hist def
				 inherit-input-method)))
    (completion-get-data-string display-string collection predicate)
    ))

(advice-add #'completing-read-default :around #'find-absfile-completing-read-default-advice)

(add-to-list 'completion-category-defaults '(absfile (styles . (uniquify-absfile))))

(add-to-list 'completion-styles-alist
	     '(uniquify-absfile
	       completion-uniquify-absfile-try-completion
	       completion-uniquify-absfile-all-completions
	       "display uniquified filenames."))

(defun find-absfile-completion-table (path string _pred action)
  "Do completion for file names in `find-absfile'.

PATH is a list of strings containing absolute directory names to
be searched.

STRING is the current user input.

PRED is ignored. Filenames matching
`completion-ignored-extensions' and filenames not matching
`completion-regexp-list' are excluded, with `case-fold-search'
bound to the value of `completion-ignore-case'.

ACTION is the current completion action; one of:

- nil; return common prefix, nil or t; see `try-completion'

- t; return all completions; see `all-completions'

- lambda; return non-nil if string is a valid completion; see
  `test-completion'.

- '(boundaries . SUFFIX); return the completion region
  '(boundaries START . END); see `completion-boundaries'.

- 'metadata; return (metadata . ALIST) as defined by
  `completion-metadata'. In addition, returns `path'
  metadata, for use by `completion-get-data-string'."

  (cond
   ((eq (car-safe action) 'boundaries)
    ;; FIXME: not doing completion on partial directory names yet
    (cons 'boundaries (cons 0 (length string))))

   ((eq action 'metadata)
    (cons 'metadata
	  (list
	   '(category . absfile)
	   (cons 'path path))))

   ((eq action 'lambda)
    ;; Called from test-completion
    (and (file-name-absolute-p string)
	 (file-exists-p string)))

   (t
    ;; `action' is one of:
    ;; t - return all completions
    ;; nil - return common prefix, nil or t

    (let ((dirname (file-name-directory string))
	  (filename (file-name-nondirectory string))
	  (result nil)) ;; list of abs filenames

      (dolist (dir path)
	;; `path' is often built by hand, so it may contain names
	;; that do not exist on the disk; check for that. It may not
	;; end in "/"; file-name-as-directory ensures that. On
	;; Windows, it may not contain the leading device (ie
	;; "c:"). `dirname' will contain the device;
	;; expand-file-name ensures that.
	(when (file-directory-p dir)
	  (setq dir (expand-file-name (file-name-as-directory dir)))

	  (when (or (= 0 (length dirname))
		    (string-equal dir dirname))
	    (setq result
		  (append result
			  (cl-mapcar (lambda (file)
				       (if action
					   (concat dir file)
					 file))
				     (file-name-all-completions filename dir))))
	    )))

      (if action
	  result

	(cond
	 ((null result)
	  nil)

	 ((and (= 1 (length result))
	       (= (length string) (length (car result))))
	  ;; unique exact match
	  t)

	 ((= 1 (length result))
	  ;; unique but not exact
	  (car result))

	 (t
	  (let ((temp (try-completion string result)))
	    (if (eq t temp)
		;; If result is all duplicates; `try-completion' returns t; return just one.
		(list (car result))
	      temp)))
	 )))
      )
    ))

(defun find-absfile (path)
  "Find a file with completion in PATH."
  (let ((table (apply-partially 'find-absfile-completion-table path)))
    (find-file (completing-read "file: " table nil t ""))
    ))

(provide 'minibuffer-patches)
